from .backends.anthropic import AnthropicLLM
from .backends.azure import AzureLLM
from .backends.custom import OpenaiCompatibleLLM
from .backends.deepseek import DeepseekLLM
from .backends.genai import GenaiLLM
from .backends.github import GithubLLM
from .backends.googleai import GoogleaiLLM
from .backends.groq import GroqLLM
from .backends.llamacpp import LlamacppLLM
from .backends.mistral import MistralLLM
from .backends.ollama import OllamaLLM
from .backends.openai import OpenaiLLM
from .backends.xai import XaiLLM

from .utils.instructions import getRagPrompt
from .utils.retrieve_text_output import getChatCompletionText
from .utils.handle_text import readTextFile

from typing import Optional, Callable, Union, Any, List, Dict
from copy import deepcopy
from io import StringIO
import sys, os, json

PACKAGE_PATH = os.path.dirname(os.path.realpath(__file__))
PACKAGE_NAME = os.path.basename(PACKAGE_PATH)
SUPPORTED_BACKENDS = ["anthropic", "azure", "custom", "deepseek", "genai", "github", "googleai", "groq", "llamacpp", "mistral", "ollama", "openai", "vertexai", "xai"]

def getChatResponse(
        messages: Union[dict, str],
        backend: str, # check SUPPORTED_BACKENDS for supported backends
        model: Optional[str]=None, # applicable to all backends, execept for llamacpp
        model_keep_alive: Optional[int]=None, # time to keep the model loaded in memory; applicable to ollama only
        system: Optional[Union[list, str]]=None, # system message; define how the model should generally behave and respond; accepts a list of strings or a single string; loop through multiple system messages for multi-turn inferences if it is a list
        context: Optional[Union[list, str]]=None, # predefined context to be added to the user prompt as prefix; accepts a list of strings or a single string; loop through multiple predefined contexts for multi-turn inferences if it is a list
        tool: Optional[Union[list, str]]=None, # a tool either a built-in tool name under the folder `tools` in the package directory or a file path of the tool; accepts a list of strings or a single string; loop through multiple tools for multi-turn actions if it is a list; parameters of both `schema` and `func` are ignored when `tool` parameter is given
        schema: Optional[dict]=None, # json schema for structured output or function calling
        func: Callable[..., Optional[str]]=None, # function to be called
        temperature: Optional[float]=None, # temperature for sampling
        max_tokens: Optional[int]=None, # maximum number of tokens to generate
        context_window: Optional[int]=None, # context window size; applicable to ollama only
        batch_size: Optional[int]=None, # batch size; applicable to ollama only
        prefill: Optional[str]=None, # prefill of assistant message; applicable to deepseek, mistral, ollama and groq only
        stop: Optional[list]=None, # stop sequences
        stream: Optional[bool]=False, # streaming events object
        stream_events_only: Optional[bool]=False, # return streaming events object only
        api_key: Optional[str]=None, # API key or credentials json file path in case of using Vertex AI as backend; applicable to anthropic, custom, deepseek, genai, github, googleai, groq, mistral, openai, xai
        api_endpoint: Optional[str]=None, # applicable to anthropic, azure, custom, llamacpp, ollama
        api_project_id: Optional[str]=None, # applicable to Vertex AI only, i.e., vertexai or genai
        api_service_location: Optional[str]=None, # applicable to Vertex AI only, i.e., vertexai or genai
        api_timeout: Optional[Union[int, float]]=None, # timeout for API request; applicable to backends, execept for ollama, genai and vertexai
        print_on_terminal: Optional[bool]=True, # print output on terminal
        word_wrap: Optional[bool]=True, # word wrap output
        **kwargs,
) -> Union[List[Dict[str, str]], Any]:
    """
    Get output from running inference with the specified backend and model.
    This function returns **either**:
    1) a list of messages of interaction between user and the AI assistant
    **or**
    2) streaming events object when stream=True AND either
        i) schema is not provided
        or
        ii) schema is provided for function calling with chat extension
    """
    if backend not in SUPPORTED_BACKENDS:
        raise ValueError(f"Backend {backend} is not supported. Supported backends are {SUPPORTED_BACKENDS}")
    # deep copy both messages and schema to ensure the original ones are unchanged
    messagesCopy = deepcopy(messages) if isinstance(messages, list) else [{"role": "system", "content": "You are a helpful AI assistant."}, {"role": "user", "content": messages}]
    schemaCopy = None if schema is None else deepcopy(schema)
    # handle given system message(s)
    if system:
        system_instruction = system.pop(0) if isinstance(system, list) else system
        # check if it is a predefined system message built-in with this SDK
        possible_system_file_path = os.path.join(PACKAGE_PATH, "systems", f"{system_instruction}.md")
        if os.path.isfile(possible_system_file_path):
            system_file_content = readTextFile(possible_system_file_path)
            if system_file_content:
                system_instruction = system_file_content
        messagesCopy = updateSystemMessage(messagesCopy, system_instruction)
    # handle given predefined context(s)
    if context:
        context_content = context.pop(0) if isinstance(context, list) else context
        # check if it is a predefined context built-in with this SDK
        possible_context_file_path = os.path.join(PACKAGE_PATH, "contexts", f"{context_content}.md")
        if os.path.isfile(possible_context_file_path):
            context_file_content = readTextFile(possible_context_file_path)
            if context_file_content:
                context_content = context_file_content
        messagesCopy[-1] = context_content + messagesCopy[-1]
    # run LLM
    if schemaCopy is not None: # structured output or function calling; allow schema to be an empty dict
        dictionary_output = {} if not schemaCopy else getDictionaryOutput(
            messagesCopy,
            schemaCopy,
            backend,
            model=model,
            model_keep_alive=model_keep_alive,
            temperature=temperature,
            max_tokens=max_tokens,
            context_window=context_window,
            batch_size=batch_size,
            prefill=prefill,
            stop=stop,
            api_key=api_key,
            api_endpoint=api_endpoint,
            api_project_id=api_project_id,
            api_service_location=api_service_location,
            api_timeout=api_timeout,
            **kwargs
        )
        if func:
            # Create a StringIO object to capture the output
            terminal_output = StringIO()
            # Redirect stdout to the StringIO object
            old_stdout = sys.stdout
            sys.stdout = terminal_output
            # placeholder for function text output
            function_text_output = ""
            try:
                # execute the function
                function_response = func() if not dictionary_output else func(**dictionary_output) # returned response can be either 1) an empty string: no chat extension 2) a non-empty string: chat extension 3) none: errors encountered in executing the function
                function_text_output = terminal_output.getvalue() # capture the function text output for function calling without chat extension
            except:
                function_response = None # due to unexpected errors encountered in executing the function; fall back to regular completion
            # Restore the original stdout
            sys.stdout = old_stdout
            # handle function response
            if function_response is None: # fall back to regular completion
                return getChatResponse(
                    messagesCopy,
                    backend,
                    model=model,
                    model_keep_alive=model_keep_alive,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    context_window=context_window,
                    batch_size=batch_size,
                    prefill=prefill,
                    stop=stop,
                    stream=stream,
                    api_key=api_key,
                    api_endpoint=api_endpoint,
                    api_project_id=api_project_id,
                    api_service_location=api_service_location,
                    api_timeout=api_timeout,
                    print_on_terminal=print_on_terminal,
                    word_wrap=word_wrap,
                    **kwargs
                )
            elif function_response: # chat extension
                return getChatResponse(
                    addContextToMessages(messagesCopy, function_response),
                    backend,
                    model=model,
                    model_keep_alive=model_keep_alive,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    context_window=context_window,
                    batch_size=batch_size,
                    prefill=prefill,
                    stop=stop,
                    stream=stream,
                    api_key=api_key,
                    api_endpoint=api_endpoint,
                    api_project_id=api_project_id,
                    api_service_location=api_service_location,
                    api_timeout=api_timeout,
                    print_on_terminal=print_on_terminal,
                    word_wrap=word_wrap,
                    **kwargs
                )
            else: # empty str; function executed successfully without chat extension
                output = function_text_output if function_text_output else "Done!"
        else: # structured output
            output = json.dumps(dictionary_output, indent=4)
        if print_on_terminal:
            print(output)
    else: # regular completion
        if backend == "anthropic":
            completion = AnthropicLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "azure":
            completion = AzureLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_endpoint=api_endpoint,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "custom":
            completion = OpenaiCompatibleLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_endpoint=api_endpoint,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "deepseek":
            completion = DeepseekLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend in ("genai", "vertexai"):
            completion = GenaiLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_project_id=api_project_id,
                api_service_location=api_service_location,
                **kwargs
            )
        elif backend == "github":
            completion = GithubLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "googleai":
            completion = GoogleaiLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "groq":
            completion = GroqLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                prefill=prefill,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "llamacpp":
            completion = LlamacppLLM.getChatCompletion(
                messagesCopy,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_endpoint=api_endpoint,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "mistral":
            completion = MistralLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                prefill=prefill,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "ollama":
            completion = OllamaLLM.getChatCompletion(             
                messagesCopy,
                model=model,
                model_keep_alive=model_keep_alive,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                context_window=context_window,
                batch_size=batch_size,
                prefill=prefill,
                stop=stop,
                stream=stream,
                api_endpoint=api_endpoint,
                **kwargs
            )
        elif backend == "openai":
            completion = OpenaiLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        elif backend == "xai":
            completion = XaiLLM.getChatCompletion(
                messagesCopy,
                model=model,
                schema=schemaCopy,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                stream=stream,
                api_key=api_key,
                api_timeout=api_timeout,
                **kwargs
            )
        if stream and stream_events_only:
            return completion
        output = getChatCompletionText(backend, completion, stream=stream, print_on_terminal=print_on_terminal, word_wrap=word_wrap)
    return output

def getDictionaryOutput(
        messages: dict,
        schema: dict,
        backend: str,
        model: Optional[str]=None,
        model_keep_alive: Optional[int]=None,
        temperature: Optional[float]=None, 
        max_tokens: Optional[int]=None,
        context_window: Optional[int]=None,
        batch_size: Optional[int]=None,
        prefill: Optional[str]=None,
        stop: Optional[list]=None,
        api_key: Optional[str]=None,
        api_endpoint: Optional[str]=None,
        api_project_id: Optional[str]=None,
        api_service_location: Optional[str]=None,
        api_timeout: Optional[Union[int, float]]=None,
        **kwargs,
) -> dict:
    if backend == "anthropic":
        return AnthropicLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "azure":
        return AzureLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_endpoint=api_endpoint,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "custom":
        return OpenaiCompatibleLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_endpoint=api_endpoint,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "deepseek":
        return DeepseekLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            prefill=prefill,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend in ("genai", "vertexai"):
        return GenaiLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_project_id=api_project_id,
            api_service_location=api_service_location,
            **kwargs
        )
    elif backend == "github":
        return GithubLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "googleai":
        return GoogleaiLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "groq":
        return GroqLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            prefill=prefill,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "llamacpp":
        return LlamacppLLM.getDictionaryOutput(
            messages,
            schema,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_endpoint=api_endpoint,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "mistral":
        return MistralLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            prefill=prefill,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "ollama":
        return OllamaLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            model_keep_alive=model_keep_alive,
            temperature=temperature,
            max_tokens=max_tokens,
            context_window=context_window,
            batch_size=batch_size,
            prefill=prefill,
            stop=stop,
            api_endpoint=api_endpoint,
            **kwargs
        )
    elif backend == "openai":
        return OpenaiLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    elif backend == "xai":
        return XaiLLM.getDictionaryOutput(
            messages,
            schema,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            api_key=api_key,
            api_timeout=api_timeout,
            **kwargs
        )
    return {}

def updateSystemMessage(messages: dict, system: str) -> dict:
    if system:
        messages[-1] = {"role": "system", "content": system}
    return messages

def addContextToMessages(messages: dict, context: str) -> dict:
    messages[-1] = [{"role": "context", "content": getRagPrompt(messages[-1].get("content", ""), context)}]
    return messages
